'''mcrypt wrapper for Python 2 and 3
Copyright (C) 2010 Michael B. Trausch <mike@trausch.us>
License: GPL v3 as published by the Free Software Foundation

The EncryptedReader module is used to read data streams which have
been encrypted with the EncryptedWriter module.  Such files have a
header, as well as padding management so that files can be decrypted
without growing in length if they happen to not be an even multiple of
the encrypter's block size.

While it is possible to use this and the EncryptedWriter to work with
data streams, it is probably not a terribly good idea to do so.  There
is absolutely no protection against corruption or tampering in this
class. The user can provide such protection for him or herself, if
necessary, or use the HashedEncryptedReader/Writer classes to generate
files where tampering and/or corruption can be detected.'''

import io
import platform
import struct
from .libmcrypt import *
from .common import ALGO_MAP, MODE_MAP, ALGO_RMAP, MODE_RMAP

class EncryptedReader(io.BufferedReader):
    '''EncryptedReader(raw, passwd) -> EncryptedReader
    A sequential, readable, block-oriented encrypted stream reader

    This is just like io.BufferedReader, but it decrypts data as it
    reads.  The 'raw' parameter is passed to the io.BufferedReader
    constructor, and the password parameter is required in order to
    create the key.  Note that unlike the EncryptedWriter, it is not
    possible to set other decryption parameters: such parameters will
    be read from the stream's header (so the stream must be at the
    correct position when it is passed to this class' constructor).

    Specifically, the algorithm, mode, salt, and IV are all read from
    the stream's header.'''
    def __init__(self, raw, password):
        pyver = platform.python_version_tuple()

        if pyver[0] == '3':
            self._parentProxy = super()
        else:
            self._parentProxy = super(io.BufferedReader, self)

        self._parentProxy.__init__(raw)

        parameters = self._read_stream_header()
        self._decrypter.set_parameters(password,
                                       parameters['salt'],
                                       parameters['iv'])
        self._decrypter.begin()

        self._block_size = self._decrypter.get_block_size()
        self._buffer = bytearray()
        self._last_block_read = False
        self._blocks_read = 0

        if len(parameters['iv']) == self._block_size:
            self._read_blocks(1)
        elif len(parameters['iv']) == 0:
            return
        else:
            raise StreamError('iv not block size nor zero?')

        return

    def close(self):
        '''close() -> None

        Closes the EncryptedReader stream.'''
        self._parentProxy.close()
        self._decrypter.end()
        return

    def get_block_size(self):
        return self._decrypter.get_block_size()

    def read(self, length):
        '''read(length) -> bytes

        Reads the specified number of bytes.  If the request results
        in a partial block having to be read, then the unrequested
        portion of the block will be put into a buffer and held for
        the next request.

        As an extension to the usual reader interface, the user can
        call .get_buffer_length() to get the length of the buffer of
        pending bytes.  As a further extension, the user can call
        .read_buffer() to drain the buffer.  It is probably best to
        read the buffer at the block size, though, since the
        underlying decrypter works block-at-a-time.'''
        (blocks, bytes) = divmod(length, self._block_size)

        if len(self._buffer) == 0:
            if bytes == 0:
                return self._read_blocks(blocks)
            else:
                ret = self._read_blocks(blocks)
                self._buffer = self._read_blocks(1)
                ret += self._buffer[:bytes]
                self._buffer = self._buffer[bytes:]
                return ret
        else:
            if len(self._buffer) == length:
                ret = bytes(self._buffer)
                self._buffer = bytearray()
                return ret
            elif length < len(buffer):
                ret = bytes(self._buffer[:length])
                self._buffer = self._buffer[length:]
                return ret
            else:
                ret = bytes(self._buffer)
                self._buffer = self._read_blocks(1)
                ret += self._buffer[:bytes]
                self._buffer = self._buffer[bytes:]
                return ret

    def _pread(self, len):
        return self._parentProxy.read(len)

    def _read_blocks(self, n):
        '''Reads blocks from the underlying encrypted file.

        If it reads the special sentinel block (a block with all
        values 0xff) then it will read the next (and presumably the
        last) block in the file, too, even if that means we read more
        blocks than the user called for, because after the sentinel
        there is one more block, which indicates the size of the
        padding (if any) at the end of the last block.  This number is
        stored in the file as a two-byte unsigned integer (just in
        case there is ever a need to handle a block size larger than
        255 bytes), in the first two bytes of the last block.  All
        other bytes of this special last block ("footer") are 0xff,
        and are reserved.'''
        pt = []
        last_block_padding = 0

        if self._last_block_read == True:
            return bytes()

        for i in range(n):
            block = self._parentProxy.read(self._block_size)
            self._blocks_read += 1
            if len(block) < self._block_size:
                msg = 'short read: stream corrupt or truncated?'
                raise McryptError(msg)
            elif block == b'\xff' * self._block_size:
                self._last_block_read = True
                footer = self._parentProxy.read(self._block_size)
                unpackstr = '!H' + ('x' * (self._block_size - 2))
                padding = struct.unpack(unpackstr, footer)[0]
                break
            else:
                pt.append(block)

        if (self._last_block_read == False) or (padding == 0):
            return self._decrypter.step(b''.join(pt))
        else:
            ptx = self._decrypter.step(b''.join(pt))

            ptx = ptx[:(-padding)]
            return ptx

    def _read_stream_header(self):
        (magic, algo, mode) = (self._pread(4), self._pread(1),
                               self._pread(1))
        salt_len = self._pread(1)
        salt = self._pread(salt_len[0])
        iv_len = self._pread(1)
        if iv_len[0] > 0:
            iv = self._pread(iv_len[0])

        self._decrypter = raw_decrypt(ALGO_RMAP[algo[0]],
                                      MODE_RMAP[mode[0]])

        blksz = self._decrypter.get_block_size()
        pad_len = blksz - (self.tell() % blksz)
        if pad_len != blksz:
            self._pread(pad_len)

        if magic != b'EW0\x00':
            raise StreamError('bad magic')

        parameters = { 'algorithm': ALGO_RMAP[algo[0]],
                       'mode': MODE_RMAP[mode[0]],
                       'salt_len': salt_len,
                       'salt': salt,
                       'iv_len': iv_len,
                       'iv': iv }

        return parameters
